/*****************************/
/*     Active unicast by     */
/*   Konstantin Kabassanov   */
/*        LIP6, Paris        */
/* konstantin@kabassanov.com */
/*****************************/

/*MBONE UNICASTv4 active / MUTLICASTv4 Gateway program.*/
/*mbone-reflector-receive-v4*/
/*This program provides a gateway service*/
/*between a unicast mbone session and a */
/*multicast mbone session.*/

/*This version of the program supports will*/
/*will provide two-way unicast/multicast gateway*/
/*services for a multiple unicast participant over*/
/*a single multicast group*/

/*This requires reading from the unicast socket*/
/*and sending over the multicast socket and */
/*visa-versa.*/

/*The address and port number for the mbone data */
/*is specified on the command line (with an even port number)*/
/*The program also performs the gateway service for the RTCP*/
/*data which is on the odd numbered port one higher*/
/*than the data*/

/*The select system call is used in order handle socket*/
/*operations asynchronously*/
/*August 19, 1999 -- TWL*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/signal.h>
#include <sys/param.h>
#include <errno.h>
#include <netdb.h>
#include <arpa/inet.h>

#define       MSGBUF_SIZE     32768

struct addr_table {
  struct in_addr address;
  int flag1;
  int flag2;
};

#define TIMEOUTSECS 300 
int ucdatatotalbytesrecv = 0;
int ucdatatotalbytessent = 0;
int ucdatarecvfromcalls = 0;
int ucdatasendtocalls = 0;
int mcdatatotalbytesrecv = 0;
int mcdatatotalbytessent = 0;
int mcdatarecvfromcalls = 0;
int mcdatasendtocalls = 0;
int ucrtcptotalbytesrecv = 0;
int ucrtcptotalbytessent = 0;
int ucrtcprecvfromcalls = 0;
int ucrtcpsendtocalls = 0;
int mcrtcptotalbytesrecv = 0;
int mcrtcptotalbytessent = 0;
int mcrtcprecvfromcalls = 0;
int mcrtcpsendtocalls = 0;

void printmembers(struct addr_table*);
int programshutdown (void);

int
main (argc, argv)
     int argc;
     char **argv;
{
  /*define variables*/
  int ucdatarecvfd;    //unicast data receive socket
  int mcdatarecvfd;    //multicast data receive socket
  int ucrtcprecvfd;    //unicast rtcp receive socket
  int mcrtcprecvfd;    //multicast rtcp receive socket
  int sendfd;    //local send socket
  u_char ttl;
  struct sockaddr_in ucdatarecvaddr;
  struct sockaddr_in ucdatasendaddr;
  struct sockaddr_in mcdataaddr;
  struct sockaddr_in ucrtcprecvaddr;
  struct sockaddr_in ucrtcpsendaddr;
  struct sockaddr_in mcrtcpaddr;
  struct sockaddr_in localsendaddr;
  struct sockaddr_in sourceaddr;
  struct ip_mreq mcdatareq;
  int ucdatarecvaddrlen;
  int mcdataaddrlen;
  int ucrtcprecvaddrlen;
  int mcrtcpaddrlen;
  int sourceaddrlen;
  char ucdatarecvbuf[MSGBUF_SIZE];
  char mcdatarecvbuf[MSGBUF_SIZE];
  char ucrtcprecvbuf[MSGBUF_SIZE];
  char mcrtcprecvbuf[MSGBUF_SIZE];
  int mcdataport;
  int ucdataport;
  int mcrtcpport;
  int ucrtcpport;
  struct in_addr multicastaddress;
  struct in_addr remoteunicastaddress;
  struct timeval tv;
  fd_set readfds;
  int maxfds;
  int nfds;
  int nr;
  int ns;
  int chksrc;
  char myhostname[MAXHOSTNAMELEN];
  char addressstring[80];
  char myhostnameipaddress[80];
  struct hostent  *host;
  int debugon = 0;
  struct addr_table *ucmember =malloc(sizeof(struct addr_table));
  int n2;
  int timeoutsecs = TIMEOUTSECS;
  char inputbuf[32];
  int timerstatus;
  struct itimerval timerval;
  
  /*check command line arguments*/
  if (argc < 6) {
    fprintf (stderr, "Usages: mbone-reflector-receive-v4 [unicast address] [unicast port] [multicast address] [multicast port] [ttl]\n");
    exit(1);
  }
  
  
  /*now read in the command line arguments*/ 
 if((inet_pton(AF_INET,argv[1],&ucmember->address.s_addr))!=1){
    printf("bad unicast address format\n");
    exit (1);
  }


  /*check if port specified on command line is even or odd*/
  /*adjust port values so that the data port is the even one*/
  /*and the rtcp port is the odd port one higher than the data port*/ 
  if (atoi(argv[2]) % 2 == 0) {
    ucdataport = atoi(argv[2]);
    ucrtcpport = atoi(argv[2]) + 1;
  } else {
    ucdataport = atoi(argv[2]) - 1;
    ucrtcpport = atoi(argv[2]);
  }
  if((inet_pton(AF_INET,argv[3],&multicastaddress.s_addr))!=1){
    printf("bad multicast address format\n");
    exit (1);
  }
  
  if (atoi(argv[4]) % 2 == 0) {
    mcdataport = atoi(argv[4]);
    mcrtcpport = atoi(argv[4]) + 1;
  } else {
    mcdataport = atoi(argv[4]) - 1;
    mcrtcpport = atoi(argv[4]);
  }
  ttl = atoi(argv[5]);
  if (argv[6] != NULL) {
    if (!strcmp(argv[6],"d")) { 
      printf("command line operation is on\n");
      debugon = 1;
    }
    if (!strcmp(argv[6],"d2")) { 
      printf("debug level 2 is on\n");
      debugon = 2;
    }
    if (!strcmp(argv[6],"d3")) { 
      printf("debug level 3 is on\n");
      debugon = 3;
    }
    if (!strcmp(argv[6],"d4")) { 
      printf("debug level 4 is on\n");
      debugon = 4;
    }
  }
 printf("unicastaddress=%s   ",argv[1]);
  printf("ucdataport=%d  ",ucdataport);
  printf("ucrtcpport=%d\n",ucrtcpport);
  printf("multicastaddress=%s   ",argv[3]);
  printf("mcdataport=%d  ",mcdataport);
  printf("mcrtcpport=%d\n",mcrtcpport);
  
  /*enter the address/port data into the ucdatarecvaddr structure*/
  bzero((char *) &ucdatarecvaddr, sizeof(ucdatarecvaddr));
  ucdatarecvaddr.sin_family = AF_INET;
  ucdatarecvaddr.sin_addr.s_addr = INADDR_ANY;
  ucdatarecvaddr.sin_port = htons(ucdataport);
  
  /*enter the address/port data into the ucdatasendaddr structure*/
  bzero((char *) &ucdatasendaddr, sizeof(ucdatasendaddr));
  ucdatasendaddr.sin_family = AF_INET;
  ucdatasendaddr.sin_addr.s_addr = remoteunicastaddress.s_addr;
  ucdatasendaddr.sin_port = htons(ucdataport);
  
  /*enter the address/port data into the ucdatarecvaddr structure*/
  bzero((char *) &ucdatarecvaddr, sizeof(ucdatarecvaddr));
  ucdatarecvaddr.sin_family = AF_INET;
  ucdatarecvaddr.sin_addr.s_addr =INADDR_ANY;
  ucdatarecvaddr.sin_port = htons(ucdataport);
  
  /*enter the address/port data into the ucdatasendaddr structure*/
  bzero((char *) &ucdatasendaddr, sizeof(ucdatasendaddr));
  ucdatasendaddr.sin_family = AF_INET;
  ucdatasendaddr.sin_addr.s_addr = remoteunicastaddress.s_addr;
  ucdatasendaddr.sin_port = htons(ucdataport);
  
  /*enter the address/port data into the ucrtcprecvaddr structure*/
  bzero((char *) &ucrtcprecvaddr, sizeof(ucrtcprecvaddr));
  ucrtcprecvaddr.sin_family = AF_INET;
  ucrtcprecvaddr.sin_addr.s_addr = INADDR_ANY;
  ucrtcprecvaddr.sin_port = htons(ucrtcpport);
  
  /*enter the address/port data into the ucrtcpsendaddr structure*/
  bzero((char *) &ucrtcpsendaddr, sizeof(ucrtcpsendaddr));
  ucrtcpsendaddr.sin_family = AF_INET;
  ucrtcpsendaddr.sin_addr.s_addr = remoteunicastaddress.s_addr;
  ucrtcpsendaddr.sin_port = htons(ucrtcpport);
  
  /*enter the address/port data into the mcdataaddr structure*/
  bzero((char *) &mcdataaddr, sizeof(mcdataaddr));
  mcdataaddr.sin_family=AF_INET;
  mcdataaddr.sin_addr.s_addr = multicastaddress.s_addr;
  mcdataaddr.sin_port = htons(mcdataport);
  
  /*enter the address/port data into the mcrtcpaddr structure*/
  bzero((char *) &mcrtcpaddr, sizeof(mcrtcpaddr));
  mcrtcpaddr.sin_family=AF_INET;
  mcrtcpaddr.sin_addr.s_addr = multicastaddress.s_addr;
  mcrtcpaddr.sin_port = htons(mcrtcpport);
  
  /*figure out the hostname and get a local ip address*/
  if (gethostname(myhostname, MAXHOSTNAMELEN) < 0)
    perror("gethostname error");
  printf("myhostname=%s\n", myhostname);
  
  if((host = gethostbyname2(myhostname,AF_INET)) == NULL) 
    perror("gethostbyname failure!");
  
  inet_ntop(AF_INET,(struct in_addr*)host->h_addr_list[0],myhostnameipaddress,sizeof(myhostnameipaddress));
  printf("myhostipaddress=%s\n\n", myhostnameipaddress);
  
  /*enter the address/port data into the localsendaddr structure*/
  bzero((char *) &localsendaddr, sizeof(localsendaddr));
  localsendaddr.sin_family=AF_INET;
  inet_pton(AF_INET, myhostnameipaddress ,&localsendaddr.sin_addr.s_addr);
  localsendaddr.sin_port = htons(0);
  
  /*enter the address/port data into the mcdatareq structure*/
  bzero((char *) &mcdatareq, sizeof(mcdatareq));
  mcdatareq.imr_multiaddr.s_addr=multicastaddress.s_addr;
  mcdatareq.imr_interface.s_addr=htonl(INADDR_ANY);
  
  /*get the ucdatarecvfd socket, bind to address*/
  if ((ucdatarecvfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open ucdatarecvfd socket!");
    exit (1);
  }
  if (bind(ucdatarecvfd, (struct sockaddr *) &ucdatarecvaddr, \
	   sizeof(ucdatarecvaddr)) < 0) {
    perror("can't bind ucdatarecvaddr to socket!");
    exit(1);
  }
  
  /*get the sendfd socket*/
  if ((sendfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open sendfd socket!");
    exit (1);
  }
  if (bind(sendfd, (struct sockaddr *) &localsendaddr, \
	   sizeof(localsendaddr)) < 0) {
    perror("can't bind localsendaddr to socket!");
    exit(1);
  }
  
  /*get the ucrtcprecvfd socket, bind to address*/
  if ((ucrtcprecvfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open ucrtcprecvfd socket!");
    exit (1);
  }
  if (bind(ucrtcprecvfd, (struct sockaddr *) &ucrtcprecvaddr, \
	   sizeof(ucrtcprecvaddr)) < 0) {
    perror("can't bind ucrtcprecvaddr to socket!");
    exit(1);
  }
  
  /*get a mcdatarecvfd socket, bind to address */
  if ((mcdatarecvfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open mcdatarecvfd socket!");
    exit(1);
  }
  if (bind(mcdatarecvfd, (struct sockaddr *) &mcdataaddr, \
	   sizeof(mcdataaddr)) < 0) {
    perror("can't bind mcdataaddr to socket!");
    exit(1);
  }
  
  /*get a mcrtcprecvfd socket, bind to address */
  if ((mcrtcprecvfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open mcrtcprecvfd socket!");
    exit(1);
  }
  if (bind(mcrtcprecvfd, (struct sockaddr *) &mcrtcpaddr, \
	   sizeof(mcrtcpaddr)) < 0) {
    perror("can't bind mcrtcpaddr to socket!");
    exit(1);
  }
  
  /*allow multiple processes per host to read from socket*/
  if (setsockopt(mcdatarecvfd, SOL_SOCKET, SO_REUSEADDR, &mcdataaddr, sizeof(mcdataaddr)) < 0) {
    perror("SO_REUSEADDR setsockopt\n");
  }  
  if (setsockopt(mcrtcprecvfd, SOL_SOCKET, SO_REUSEADDR, &mcrtcpaddr, sizeof(mcrtcpaddr)) < 0) {
    perror("SO_REUSEADDR setsockopt\n");
  }  
  
  /*set socket options to join multicast group*/
  /*setsockopt (int filedescriptor, int level, int optname,*/  
  /*            charpointer optvalue, intpointer optlen)*/
  if (setsockopt(mcdatarecvfd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mcdatareq,\
		 sizeof(mcdatareq)) < 0) {
    perror("can't set socket options to join multicast group!");
    exit(1);
  } 
  if (setsockopt(mcrtcprecvfd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mcdatareq,\
		 sizeof(mcdatareq)) < 0) {
    perror("can't set socket options to join multicast group!");
    exit(1);
  } 
  
  /*now set multicast socket TTL option*/
  /*setsocketopt (int filedescriptor, int level, in optname),*/
  /*charpointer optvalue, intpointer optlen)*/
    if (setsockopt(sendfd, IPPROTO_IP, IP_MULTICAST_TTL, \
		 &ttl, sizeof(ttl)) < 0)
    perror("can't set multicast ttl socket option!");
  
  /*define the sizes of the address structures*/
  ucdatarecvaddrlen=sizeof(ucdatarecvaddr);
  mcdataaddrlen=sizeof(mcdataaddr);
  ucrtcprecvaddrlen=sizeof(ucrtcprecvaddr);
  mcrtcpaddrlen=sizeof(mcrtcpaddr);
  sourceaddrlen=sizeof(sourceaddr);
  
  /*set up timer timeout values, call setitimer*/
  /*set interval to 0, will reset timer as part of timeout processing*/
  timerval.it_interval.tv_sec = 0; 
  timerval.it_interval.tv_usec = 0; 
  timerval.it_value.tv_sec = timeoutsecs; 
  timerval.it_value.tv_usec = 0; 
  /*ignore SIGALRM*/
  if (signal(SIGALRM,SIG_IGN) != SIG_IGN)
    timerstatus = setitimer(ITIMER_REAL, &timerval, (struct itimerval *)0);
  
  
  /*start infinite while loop*/
  /*and check for activity on sockets*/
  
  /*set up the select sytem call parameters*/  
  /*zero out the readfds and writefds lists, then add*/
  /*the file descriptors of interest*/
  while(1) { 
    FD_ZERO(&readfds);
    FD_SET(ucdatarecvfd, &readfds);
    FD_SET(mcdatarecvfd, &readfds);
    FD_SET(ucrtcprecvfd, &readfds);
    FD_SET(mcrtcprecvfd, &readfds);
    if (debugon >= 1) {
      FD_SET(0, &readfds);
    }
    /*set time value to 0 for continuous polling*/
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    maxfds = mcdatarecvfd +10;
    
    /*check for activity and processor accordingly*/
    nfds = select(maxfds, &readfds, NULL, NULL, NULL);
    
    /*if specified on the command line, check for input on stdin*/
    if (debugon >= 1) {
      if (FD_ISSET(0, &readfds)) {
	n2 = read(0,inputbuf, sizeof(inputbuf));
	inputbuf[n2>0? n2-1: 0] = '\0';
	if (!strcmp(inputbuf,"p")) {
	  printf("current unicast members:\n");
	  printmembers(ucmember);
	} 
	if (!strcmp(inputbuf,"q")) {
	  programshutdown();
	} 
      }
      fflush(stdout); 
    }
    
    
    /*1:receive from unicast, send on multicast  - data*/ 
    if (FD_ISSET(ucdatarecvfd, &readfds)) {
      nr = recvfrom(ucdatarecvfd, ucdatarecvbuf, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr, &sourceaddrlen);
      if (debugon >= 2){ 
	inet_ntop(AF_INET,&sourceaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("\nreading from ucdatarecvfd, got data from %s\n", addressstring);
      }
      ucdatarecvfromcalls = ucdatarecvfromcalls++;
      ucdatatotalbytesrecv = ucdatatotalbytesrecv + nr;
      if (nr < 0)
	printf ("ucdatarecvfd:recvfrom over unicast address error!(1)\n");     
      
      /*sent to the multicast group*/
      if (debugon >= 2){
	inet_ntop(AF_INET,&mcdataaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("sending to %s\n", addressstring);
      }
      ns = sendto(sendfd, ucdatarecvbuf, nr, 0, (struct sockaddr *)&mcdataaddr, \
		  sizeof(mcdataaddr));
      mcdatasendtocalls = mcdatasendtocalls++;
      mcdatatotalbytessent = mcdatatotalbytessent + ns;
      if (ns < 0)
	printf("sendfd:sendto over multicast address error!(2)\n");
    }
    
    /*2:receive from unicast, send on multicast  - rtcp*/
    if (FD_ISSET(ucrtcprecvfd, &readfds)) {
      nr = recvfrom(ucrtcprecvfd, ucrtcprecvbuf, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr, &sourceaddrlen);
      if (debugon >= 2) {
	inet_ntop(AF_INET,&sourceaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("\nreading from ucrtcprecvfd, got data from %s\n", addressstring);
      }
      ucrtcprecvfromcalls = ucrtcprecvfromcalls++;
      ucrtcptotalbytesrecv = ucrtcptotalbytesrecv + nr; 
      if (nr < 0)
	printf ("ucrtcprecvfd:recvfrom over unicast address error!(3)\n");
      
      /*sent to the multicast group*/
      if (debugon >= 2) {
	inet_ntop(AF_INET,&mcrtcpaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("sending to %s\n", addressstring);
      }
      ns = sendto(sendfd, ucrtcprecvbuf, nr, 0, (struct sockaddr *)&mcrtcpaddr, \
		  sizeof(mcrtcpaddr));
      mcrtcpsendtocalls = mcrtcpsendtocalls++;
      mcrtcptotalbytessent = mcrtcptotalbytessent + ns;
      if (ns < 0)
	printf("sendfd:sendto over multicast address error!(4)\n");
    }
    
    /*3:receive from multicast, send on unicast to unicast participant - data*/ 
    if (FD_ISSET(mcdatarecvfd, &readfds)) {
      nr = recvfrom(mcdatarecvfd, mcdatarecvbuf, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr, &sourceaddrlen);
      if (debugon >= 2)  {
	inet_ntop(AF_INET,&sourceaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("\nreading from mcdatarecvfd, got data from=%s\n",addressstring);
	inet_ntop(AF_INET,&localsendaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("localsendaddr=%s\n",addressstring);
      }
      if (sourceaddr.sin_addr.s_addr== localsendaddr.sin_addr.s_addr
){
	chksrc = 0;
	if (debugon >= 2) 
	  printf("don't retransmit multicast sourced from gateway machine\n");
      } else {
	chksrc = 1;
	if (debugon >= 2) 
	  printf("retransmit to unicast address\n");
      }
      
      if (chksrc != 0 ) {
	mcdatarecvfromcalls = mcdatarecvfromcalls++;
	mcdatatotalbytesrecv = mcdatatotalbytesrecv + nr;
	if (nr < 0)
	  printf ("mcdatarecvfd:recvfrom over multicast address error!(5)\n");
	
	/*now send to unicast member*/
	  remoteunicastaddress.s_addr = ucmember->address.s_addr;
	  if (remoteunicastaddress.s_addr!=INADDR_ANY) {
	    ucdatasendaddr.sin_addr.s_addr = remoteunicastaddress.s_addr;
	    if (debugon >= 2) {
	      inet_ntop(AF_INET,&ucmember->address.s_addr,addressstring ,sizeof(addressstring));
	      printf("sending to %s\n", addressstring);
	    }
	    ns = sendto(sendfd, mcdatarecvbuf, nr, 0, (struct sockaddr *)&ucdatasendaddr, \
			sizeof(ucdatasendaddr));
	    ucdatasendtocalls = ucdatasendtocalls++;
	    ucdatatotalbytessent = ucrtcptotalbytessent++;
	  } else {
	    if (debugon >= 2) printf("not resending to ORIGINATOR! or array entry = 0\n");
	  }
      }//end of if (chksrc == 0)
    }
    
    /*4:receive from multicast, send on unicast - rtcp*/ 
    if (FD_ISSET(mcrtcprecvfd, &readfds)) {
      nr = recvfrom(mcrtcprecvfd, mcrtcprecvbuf, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr, &sourceaddrlen);
      if (debugon >= 2)  {
	inet_ntop(AF_INET,&sourceaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("\nreading from mcrtcprecvfd, got data from=%s\n",addressstring);
	inet_ntop(AF_INET,&localsendaddr.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("localsendaddr=%s\n",addressstring);
      }
      if (sourceaddr.sin_addr.s_addr==localsendaddr.sin_addr.s_addr){
	chksrc = 0;
	if (debugon >= 2)
	  printf("don't retransmit\n");
      } else {
	chksrc = 1;
	if (debugon >= 2)
	  printf("retransmit to unicast address\n");
      }
      if (chksrc != 0 ) {
	mcrtcprecvfromcalls = mcrtcprecvfromcalls++;
	mcrtcptotalbytesrecv = mcrtcptotalbytesrecv + nr;
	if (nr < 0)
	  printf ("mcrtcprecvfd:recvfrom over multicast address error!(6)\n");
	
	/*now  send to unicast member*/
	  remoteunicastaddress.s_addr = ucmember->address.s_addr;
	  if (remoteunicastaddress.s_addr!=INADDR_ANY)
 {
	    ucrtcpsendaddr.sin_addr.s_addr = remoteunicastaddress.s_addr;
	    if (debugon >= 2){
	      inet_ntop(AF_INET,&ucmember->address.s_addr,addressstring ,sizeof(addressstring));
	      printf("sending to %s\n", addressstring);
	    }
	    ns = sendto(sendfd, mcrtcprecvbuf, nr, 0, (struct sockaddr *)&ucrtcpsendaddr, \
			sizeof(ucrtcpsendaddr));
	    ucrtcpsendtocalls = ucrtcpsendtocalls++;
	    ucrtcptotalbytessent = ucrtcptotalbytessent++;
	  } else {
	    if (debugon >= 4) printf("not resending to ORIGINATOR! or array entry = 0\n");
	  }
      }//end of if (chksrc == 0)
    }
    
  } //end of while loop
} //end of main


int programshutdown () {
extern int ucdatatotalbytesrecv;
extern int mcdatatotalbytesrecv;
extern int ucdatarecvfromcalls;
extern int mcdatarecvfromcalls;
extern int ucdatatotalbytessent;
extern int mcdatatotalbytessent;
extern int ucdatasendtocalls;
extern int mcdatasendtocalls;
extern int ucrtcptotalbytesrecv;
extern int mcrtcptotalbytesrecv;
extern int ucrtcprecvfromcalls;
extern int mcrtcprecvfromcalls;
extern int ucrtcptotalbytessent;
extern int mcrtcptotalbytessent;
extern int ucrtcpsendtocalls;
extern int mcrtcpsendtocalls;
printf ("\n");

/*print out debug information*/
printf("number of uc data bytes received = %d\n", ucdatatotalbytesrecv);
printf("number of uc data recvfromcalls = %d\n", ucdatarecvfromcalls);
printf("number of mc data bytes sent = %d\n", mcdatatotalbytessent);
printf("number of mc data sendtocalls = %d\n", mcdatasendtocalls);
printf("--------------------------------\n");
printf("number of uc rtcp bytes received = %d\n", ucrtcptotalbytesrecv);
printf("number of uc rtcp recvfromcalls = %d\n", ucrtcprecvfromcalls);
printf("number of mc rtcp bytes sent = %d\n", mcrtcptotalbytessent);
printf("number of mc rtcp sendtocalls = %d\n", mcrtcpsendtocalls);
printf("--------------------------------\n");
printf("number of mc data bytes received = %d\n", mcdatatotalbytesrecv);
printf("number of mc data recvfromcalls = %d\n", mcdatarecvfromcalls);
printf("number of uc data bytes sent = %d\n", ucdatatotalbytessent);
printf("number of uc data sendtocalls = %d\n", ucdatasendtocalls);
printf("--------------------------------\n");
printf("number of mc rtcp bytes received = %d\n", mcrtcptotalbytesrecv);
printf("number of mc rtcp recvfromcalls = %d\n", mcrtcprecvfromcalls);
printf("number of uc rtcp bytes sent = %d\n", ucrtcptotalbytessent);
printf("number of uc rtcp sendtocalls = %d\n", ucrtcpsendtocalls);
exit(0);
}
 

void printmembers(struct addr_table* array)  {
 char addressstring[80];

 if (array->address.s_addr!=INADDR_ANY) {
         //printf("%d\n",array[stopcond][0]);
   inet_ntop(AF_INET,&array->address.s_addr,addressstring ,sizeof(addressstring));
         printf("%s  ",addressstring);
 }     
}



